from unittest import TestCase

import inspect
import shutil
import tempfile
from mock import Mock, MagicMock
from random import randrange

from autosubmit.job.job import Job
from autosubmit.job.job_common import Status
from autosubmit.job.job_list import JobList
from autosubmit.job.job_list_persistence import JobListPersistenceDb
from autosubmitconfigparser.config.yamlparser import YAMLParserFactory


class TestJobList(TestCase):
    def setUp(self):
        self.experiment_id = 'random-id'
        self.as_conf = Mock()
        self.as_conf.experiment_data = dict()
        self.as_conf.experiment_data["JOBS"] = dict()
        self.as_conf.jobs_data = self.as_conf.experiment_data["JOBS"]
        self.as_conf.experiment_data["PLATFORMS"] = dict()
        self.temp_directory = tempfile.mkdtemp()
        self.job_list = JobList(self.experiment_id, FakeBasicConfig, YAMLParserFactory(),
                                JobListPersistenceDb(self.temp_directory, 'db'), self.as_conf)
        dummy_serial_platform = MagicMock()
        dummy_serial_platform.name = 'serial'
        dummy_platform = MagicMock()
        dummy_platform.serial_platform = dummy_serial_platform
        dummy_platform.name = 'dummy_platform'
        # creating jobs for self list
        self.completed_job = self._createDummyJobWithStatus(Status.COMPLETED)
        self.completed_job.platform = dummy_platform
        self.completed_job2 = self._createDummyJobWithStatus(Status.COMPLETED)
        self.completed_job2.platform = dummy_platform
        self.completed_job3 = self._createDummyJobWithStatus(Status.COMPLETED)
        self.completed_job3.platform = dummy_platform
        self.completed_job4 = self._createDummyJobWithStatus(Status.COMPLETED)
        self.completed_job4.platform = dummy_platform
        self.submitted_job = self._createDummyJobWithStatus(Status.SUBMITTED)
        self.submitted_job.platform = dummy_platform
        self.submitted_job2 = self._createDummyJobWithStatus(Status.SUBMITTED)
        self.submitted_job2.platform = dummy_platform
        self.submitted_job3 = self._createDummyJobWithStatus(Status.SUBMITTED)
        self.submitted_job3.platform = dummy_platform

        self.running_job = self._createDummyJobWithStatus(Status.RUNNING)
        self.running_job.platform = dummy_platform
        self.running_job2 = self._createDummyJobWithStatus(Status.RUNNING)
        self.running_job2.platform = dummy_platform

        self.queuing_job = self._createDummyJobWithStatus(Status.QUEUING)
        self.queuing_job.platform = dummy_platform

        self.failed_job = self._createDummyJobWithStatus(Status.FAILED)
        self.failed_job.platform = dummy_platform
        self.failed_job2 = self._createDummyJobWithStatus(Status.FAILED)
        self.failed_job2.platform = dummy_platform
        self.failed_job3 = self._createDummyJobWithStatus(Status.FAILED)
        self.failed_job3.platform = dummy_platform
        self.failed_job4 = self._createDummyJobWithStatus(Status.FAILED)
        self.failed_job4.platform = dummy_platform
        self.ready_job = self._createDummyJobWithStatus(Status.READY)
        self.ready_job.platform = dummy_platform
        self.ready_job2 = self._createDummyJobWithStatus(Status.READY)
        self.ready_job2.platform = dummy_platform
        self.ready_job3 = self._createDummyJobWithStatus(Status.READY)
        self.ready_job3.platform = dummy_platform

        self.waiting_job = self._createDummyJobWithStatus(Status.WAITING)
        self.waiting_job.platform = dummy_platform
        self.waiting_job2 = self._createDummyJobWithStatus(Status.WAITING)
        self.waiting_job2.platform = dummy_platform

        self.unknown_job = self._createDummyJobWithStatus(Status.UNKNOWN)
        self.unknown_job.platform = dummy_platform


        self.job_list._job_list = [self.completed_job, self.completed_job2, self.completed_job3, self.completed_job4,
                                   self.submitted_job, self.submitted_job2, self.submitted_job3, self.running_job,
                                   self.running_job2, self.queuing_job, self.failed_job, self.failed_job2,
                                   self.failed_job3, self.failed_job4, self.ready_job, self.ready_job2,
                                   self.ready_job3, self.waiting_job, self.waiting_job2, self.unknown_job]
        self.waiting_job.parents.add(self.ready_job)
        self.waiting_job.parents.add(self.completed_job)
        self.waiting_job.parents.add(self.failed_job)
        self.waiting_job.parents.add(self.submitted_job)
        self.waiting_job.parents.add(self.running_job)
        self.waiting_job.parents.add(self.queuing_job)

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_directory)

    def test_add_edge_job(self):
        special_variables = dict()
        special_variables["STATUS"] = Status.VALUE_TO_KEY[Status.COMPLETED]
        special_variables["FROM_STEP"] = 0
        for p in self.waiting_job.parents:
            self.waiting_job.add_edge_info(p, special_variables)
        for parent in self.waiting_job.parents:
            self.assertEqual(self.waiting_job.edge_info[special_variables["STATUS"]][parent.name],
                             (parent, special_variables.get("FROM_STEP", 0)))


    def test_add_edge_info_joblist(self):
        special_conditions = dict()
        special_conditions["STATUS"] = Status.VALUE_TO_KEY[Status.COMPLETED]
        special_conditions["FROM_STEP"] = 0
        self.job_list._add_edge_info(self.waiting_job, special_conditions["STATUS"])
        self.assertEqual(len(self.job_list.jobs_edges.get(Status.VALUE_TO_KEY[Status.COMPLETED],[])),1)
        self.job_list._add_edge_info(self.waiting_job2, special_conditions["STATUS"])
        self.assertEqual(len(self.job_list.jobs_edges.get(Status.VALUE_TO_KEY[Status.COMPLETED],[])),2)

    def test_check_special_status(self):
        self.waiting_job.edge_info = dict()

        self.job_list.jobs_edges = dict()
        # Adds edge info for waiting_job in the list
        self.job_list._add_edge_info(self.waiting_job, Status.VALUE_TO_KEY[Status.COMPLETED])
        self.job_list._add_edge_info(self.waiting_job, Status.VALUE_TO_KEY[Status.READY])
        self.job_list._add_edge_info(self.waiting_job, Status.VALUE_TO_KEY[Status.RUNNING])
        self.job_list._add_edge_info(self.waiting_job, Status.VALUE_TO_KEY[Status.SUBMITTED])
        self.job_list._add_edge_info(self.waiting_job, Status.VALUE_TO_KEY[Status.QUEUING])
        self.job_list._add_edge_info(self.waiting_job, Status.VALUE_TO_KEY[Status.FAILED])
        # Adds edge info for waiting_job
        special_variables = dict()
        for p in self.waiting_job.parents:
            special_variables["STATUS"] = Status.VALUE_TO_KEY[p.status]
            special_variables["FROM_STEP"] = 0
            self.waiting_job.add_edge_info(p,special_variables)
        # call to special status
        jobs_to_check = self.job_list.check_special_status()
        for job in jobs_to_check:
            tmp = [parent for parent in job.parents if
                   parent.status == Status.COMPLETED or parent in self.jobs_edges["ALL"]]
            assert len(tmp) == len(job.parents)
        self.waiting_job.add_parent(self.waiting_job2)
        for job in jobs_to_check:
            tmp = [parent for parent in job.parents if
                   parent.status == Status.COMPLETED or parent in self.jobs_edges["ALL"]]
            assert len(tmp) == len(job.parents)



    def _createDummyJobWithStatus(self, status):
        job_name = str(randrange(999999, 999999999))
        job_id = randrange(1, 999)
        job = Job(job_name, job_id, status, 0)
        job.type = randrange(0, 2)
        return job

class FakeBasicConfig:
    def __init__(self):
        pass
    def props(self):
        pr = {}
        for name in dir(self):
            value = getattr(self, name)
            if not name.startswith('__') and not inspect.ismethod(value) and not inspect.isfunction(value):
                pr[name] = value
        return pr
    DB_DIR = '/dummy/db/dir'
    DB_FILE = '/dummy/db/file'
    DB_PATH = '/dummy/db/path'
    LOCAL_ROOT_DIR = '/dummy/local/root/dir'
    LOCAL_TMP_DIR = '/dummy/local/temp/dir'
    LOCAL_PROJ_DIR = '/dummy/local/proj/dir'
    DEFAULT_PLATFORMS_CONF = ''
    DEFAULT_JOBS_CONF = ''



#!/usr/bin/env python3

# Copyright 2015-2020 Earth Sciences Department, BSC-CNS
# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

import sqlite3
import os
import autosubmit.history.utils as HUtils
import autosubmit.history.database_managers.database_models as Models
from abc import ABCMeta

DEFAULT_JOBDATA_DIR = os.path.join('/esarchive', 'autosubmit', 'as_metadata', 'data')
DEFAULT_HISTORICAL_LOGS_DIR = os.path.join('/esarchive', 'autosubmit', 'as_metadata', 'logs')
DEFAULT_LOCAL_ROOT_DIR = os.path.join('/esarchive', 'autosubmit')
class DatabaseManager(metaclass=ABCMeta):
  """ Simple database manager. Needs expid. """
  AS_TIMES_DB_NAME = "as_times.db" # default AS_TIMES location
  ECEARTH_DB_NAME = "ecearth.db" # default EC_EARTH_DB_NAME location
  def __init__(self, expid, jobdata_dir_path=DEFAULT_JOBDATA_DIR, local_root_dir_path=DEFAULT_LOCAL_ROOT_DIR):    
    self.expid = expid
    self.JOBDATA_DIR = jobdata_dir_path
    self.LOCAL_ROOT_DIR = local_root_dir_path       

  def get_connection(self, path):
    # type : (str) -> Sqlite3Connection
    """ 
    Create a database connection to the SQLite database specified by path.  
    :param path: database file name  
    :return: Connection object or None
    """
    if not os.path.exists(path):
        self._create_database_file(path)
    return sqlite3.connect(path)  

  def _create_database_file(self, path):
    # type : (str) -> None
    """ creates a database files with full permissions """
    os.umask(0)
    os.open(path, os.O_WRONLY | os.O_CREAT, 0o776)

  def execute_statement_on_dbfile(self, path, statement):
    # type : (str, str) -> None
    """ Executes a statement on a database file specified by path. """
    conn = self.get_connection(path)
    cursor = conn.cursor()
    cursor.execute(statement)
    conn.commit()
    conn.close()        
  
  def execute_statement_with_arguments_on_dbfile(self, path, statement, arguments):
    # type : (str, str, Tuple) -> None
    """ Executes a statement with arguments on a database file specified by path. """
    conn = self.get_connection(path)
    cursor = conn.cursor()
    cursor.execute(statement, arguments)
    conn.commit()
    conn.close()

  def execute_many_statement_with_arguments_on_dbfile(self, path, statement, arguments_list):
    # type : (str, str, List[Tuple]) -> None
    """ Executes many statements from a list of arguments specified by a path. """
    conn = self.get_connection(path)
    cursor = conn.cursor()
    cursor.executemany(statement, arguments_list)
    conn.commit()
    conn.close()

  def execute_many_statements_on_dbfile(self, path, statements):
    # type : (str, List[str]) -> None
    """
    Updates the table schema using a **small** list of statements. No Exception raised.  
    Should be used to execute a list of schema updates that might have been already applied.
    """
    for statement in statements:
      try:                   
          self.execute_statement_on_dbfile(path, statement)
      except Exception as exp:      
          pass

  def get_from_statement(self, path, statement):
    # type : (str, str) -> List[Tuple]
    """ Get the rows from a statement with no arguments """
    conn = self.get_connection(path)
    conn.text_factory = str
    cursor = conn.cursor()
    cursor.execute(statement)
    statement_rows = cursor.fetchall()
    conn.close()
    return statement_rows
  
  def get_from_statement_with_arguments(self, path, statement, arguments):
    # type : (str, str, Tuple) -> List[Tuple]
    """ Get the rows from a statement with arguments """    
    conn = self.get_connection(path)
    conn.text_factory = str
    cursor = conn.cursor()
    cursor.execute(statement, arguments)
    statement_rows = cursor.fetchall()
    conn.close()
    return statement_rows
  
  def insert_statement(self, path, statement):
    # type : (str, str) -> int
    """ Insert statement into path """
    conn = self.get_connection(path)
    conn.text_factory = str
    cursor = conn.cursor()
    cursor.execute(statement)
    lastrow_id = cursor.lastrowid
    conn.commit()
    conn.close()
    return lastrow_id
  
  def insert_statement_with_arguments(self, path, statement, arguments):
    # type : (str, str, Tuple) -> int
    """ Insert statement with arguments into path """
    conn = self.get_connection(path)
    conn.text_factory = str
    cursor = conn.cursor()
    cursor.execute(statement, arguments) 
    lastrow_id = cursor.lastrowid
    conn.commit()
    conn.close()
    return lastrow_id
  
  def get_built_select_statement(self, table_name, conditions=None):
    # type : (str, namedtuple, str) -> str
    """ Build and return a SELECT statement with the same fields as the model. Requires that the table is associated with a model (namedtuple). """
    model = Models.table_name_to_model[table_name]
    if conditions:
      return "SELECT {0} FROM {1} WHERE {2}".format(HUtils.get_fields_as_comma_str(model), table_name, conditions)
    else:
      return "SELECT {0} FROM {1}".format(HUtils.get_fields_as_comma_str(model), table_name)





#!/bin/env/python

class StatsSummary(object):

  def __init__(self):    
    # Counters
    self.submitted_count = 0 
    self.run_count = 0
    self.completed_count = 0
    self.failed_count = 0
    # Consumption
    self.expected_consumption = 0.0
    self.real_consumption = 0.0
    self.failed_real_consumption = 0.0
    # CPU Consumption
    self.expected_cpu_consumption = 0.0 
    self.cpu_consumption = 0.0
    self.failed_cpu_consumption = 0.0
    self.total_queue_time = 0.0
    self.cpu_consumption_percentage = 0.0
  
  def calculate_consumption_percentage(self):
    if self.expected_cpu_consumption > 0.0:
      self.cpu_consumption_percentage = round((self.cpu_consumption / self.expected_cpu_consumption) * 100, 2)
  
  def get_as_list(self):
   return [
      "Summary: ",
      "{}  :  {}".format("CPU Consumption Percentage", str(self.cpu_consumption_percentage) + "%"),
      "{}  :  {:,} hrs.".format("Total Queue Time", round(self.total_queue_time, 2)),
      "{}  :  {:,}".format("Submitted Count", self.submitted_count),
      "{}  :  {:,}".format("Run Count", self.run_count),
      "{}  :  {:,}".format("Completed Count", self.completed_count),
      "{}  :  {:,}".format("Failed Count", self.failed_count),
      "{}  :  {:,} hrs.".format("Expected Consumption", round(self.expected_consumption, 4)),
      "{}  :  {:,} hrs.".format("Real Consumption", round(self.real_consumption, 4)),
      "{}  :  {:,} hrs.".format("Failed Real Consumption", round(self.failed_real_consumption, 4)),
      "{}  :  {:,} hrs.".format("Expected CPU Consumption", round(self.expected_cpu_consumption, 4)),
      "{}  :  {:,} hrs.".format("CPU Consumption", round(self.cpu_consumption, 4)),
      "{}  :  {:,} hrs.".format("Failed CPU Consumption", round(self.failed_cpu_consumption, 4))
    ]
  









  


#!/usr/bin/env python3

# Copyright 2014 Climate Forecasting Unit, IC3

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

from os import path
from setuptools import setup
from setuptools import find_packages

here = path.abspath(path.dirname(__file__))

# Get the version number from the relevant file
with open(path.join(here, 'VERSION')) as f:
    version = f.read().strip()

setup(
    name='autosubmit',
    license='GNU GPL v3',
    platforms=['GNU/Linux Debian'],
    version=version,
    description='Autosubmit is a Python-based workflow manager to create, manage and monitor complex tasks involving different substeps, such as scientific computational experiments. These workflows may involve multiple computing systems for their completion, from HPCs to post-processing clusters or workstations. Autosubmit can orchestrate all the tasks integrating the workflow by managing their dependencies, interfacing with all the platforms involved, and handling eventual errors.',
    long_description=open('README_PIP.md').read(),
    author='Daniel Beltran Mora',
    author_email='daniel.beltran@bsc.es',
    url='http://www.bsc.es/projects/earthscience/autosubmit/',
    download_url='https://earth.bsc.es/wiki/doku.php?id=tools:autosubmit',
    keywords=['climate', 'weather', 'workflow', 'HPC'],
    install_requires=['ruamel.yaml==0.17.21','cython','autosubmitconfigparser','bcrypt>=3.2','packaging>19','six>=1.10.0','configobj>=5.0.6','argparse>=1.4.0','python-dateutil>=2.8.2','matplotlib<3.6','numpy<1.22','py3dotplus>=1.1.0','pyparsing>=3.0.7','paramiko>=2.9.2','mock>=4.0.3','portalocker>=2.3.2','networkx==2.6.3','requests>=2.27.1','bscearth.utils>=0.5.2','cryptography>=36.0.1','setuptools>=60.8.2','xlib>=0.21','pip>=22.0.3','pythondialog','pytest','nose','coverage','PyNaCl>=1.4.0','Pygments','psutil'],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: POSIX :: Linux",
    ],
    packages=find_packages(),
    include_package_data=True,
    package_data={'autosubmit': [
        'autosubmit/config/files/autosubmit.conf',
        'autosubmit/config/files/expdef.conf',
        'autosubmit/database/data/autosubmit.sql',
        'README',
        'CHANGELOG',
        'VERSION',
        'LICENSE',
        'docs/autosubmit.pdf'
    ]
    },
    scripts=['bin/autosubmit']
)



#!/usr/bin/env python3

# Copyright 2017-2020 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.
import datetime
import textwrap


class Status:
    """
    Class to handle the status of a job
    """
    WAITING = 0
    READY = 1
    SUBMITTED = 2
    QUEUING = 3
    RUNNING = 4
    COMPLETED = 5
    HELD = 6
    PREPARED = 7
    SKIPPED = 8
    FAILED = -1
    DELAYED = 9
    UNKNOWN = -2
    SUSPENDED = -3
    #######
    # Note: any change on constants must be applied on the dict below!!!
    VALUE_TO_KEY = {-3: 'SUSPENDED', -2: 'UNKNOWN', -1: 'FAILED', 0: 'WAITING', 1: 'READY',
                    2: 'SUBMITTED', 3: 'QUEUING', 4: 'RUNNING', 5: 'COMPLETED', 6: 'HELD', 7: 'PREPARED', 8: 'SKIPPED', 9: 'DELAYED'}
    LOGICAL_ORDER = ["WAITING", "DELAYED", "PREPARED", "READY", "SUBMITTED", "HELD", "QUEUING", "RUNNING", "SKIPPED", "FAILED", "UNKNOWN", "COMPLETED", "SUSPENDED"]

    def retval(self, value):
        return getattr(self, value)


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    # Status Colors
    UNKNOWN = '\033[37;1m'
    WAITING = '\033[37m'

    READY = '\033[36;1m'
    SUBMITTED = '\033[36m'
    QUEUING = '\033[35;1m'
    RUNNING = '\033[32m'
    COMPLETED = '\033[33m'
    SKIPPED = '\033[33m'
    PREPARED = '\033[34;2m'
    HELD = '\033[34;1m'
    FAILED = '\033[31m'
    DELAYED = '\033[36;1m'
    SUSPENDED = '\033[31;1m'
    CODE_TO_COLOR = {-3: SUSPENDED, -2: UNKNOWN, -1: FAILED, 0: WAITING, 1: READY,
                     2: SUBMITTED, 3: QUEUING, 4: RUNNING, 5: COMPLETED, 6: HELD, 7: PREPARED, 8: SKIPPED, 9: DELAYED}


class Type:
    """
    Class to handle the status of a job
    """
    BASH = 0
    PYTHON = 1
    R = 2
    PYTHON2 = 3
    PYTHON3 = 4

    def retval(self, value):
        return getattr(self, value)


# TODO: Statistics classes refactor proposal: replace tailer by footer


class StatisticsSnippetBash:
    """
    Class to handle the statistics snippet of a job. It contains header and tailer for
    local and remote jobs
    """

    @staticmethod
    def as_header(scheduler_header, executable):
        if not executable:
            executable = "/bin/bash"
        return textwrap.dedent("""\
            #!{0}

            """).format(executable) + \
            scheduler_header + \
            textwrap.dedent("""\
            ###################
            # Autosubmit header
            ###################
            locale_to_set=$(locale -a | grep ^C.)
            if [ -z "$locale_to_set" ] ; then
                # locale installed...
                export LC_ALL=$locale_to_set
            else
                # locale not installed...
                locale_to_set=$(locale -a | grep ^en_GB.utf8)
                if [ -z "$locale_to_set" ] ; then
                    export LC_ALL=$locale_to_set
                else
                    export LC_ALL=C
                fi 
            fi
            
            set -xuve
            job_name_ptrn='%CURRENT_LOGDIR%/%JOBNAME%'
            echo $(date +%s) > ${job_name_ptrn}_STAT

            ################### 
            # AS CHECKPOINT FUNCTION
            ###################
            # Creates a new checkpoint file upon call based on the current numbers of calls to the function
            
            AS_CHECKPOINT_CALLS=0
            function as_checkpoint {
                AS_CHECKPOINT_CALLS=$((AS_CHECKPOINT_CALLS+1))
                touch ${job_name_ptrn}_CHECKPOINT_${AS_CHECKPOINT_CALLS}
            }
            ###################
            # Autosubmit job
            ###################

            """)

    @staticmethod
    def as_tailer():
        return textwrap.dedent("""\

                ###################
                # Autosubmit tailer
                ###################
                set -xuve
                echo $(date +%s) >> ${job_name_ptrn}_STAT
                touch ${job_name_ptrn}_COMPLETED
                exit 0

                """)


class StatisticsSnippetPython:
    """
    Class to handle the statistics snippet of a job. It contains header and tailer for
    local and remote jobs
    """
    def __init__(self,version="3"):
        self.version=version

    def as_header(self,scheduler_header, executable):
        if not executable:
            executable = "/usr/bin/env python"+str(self.version)
        else:
            executable = executable
        return textwrap.dedent("""\
            #!{0}

            """).format(executable) + \
            scheduler_header + \
            textwrap.dedent("""\
            ###################
            # Autosubmit header
            ###################
            import locale
            import time
            try:
                try:
                    locale.setlocale(locale.LC_ALL,'C.utf8')
                except Exception as e:
                    try:
                        locale.setlocale(locale.LC_ALL, 'C.UTF-8')
                    except Exception as e:
                        try:
                            locale.setlocale(locale.LC_ALL, 'en_GB')
                        except Exception as e:
                            locale.setlocale(locale.LC_ALL, 'es_ES')
            except Exception as e:
                locale.setlocale(locale.LC_ALL, 'C')
            job_name_ptrn = '%CURRENT_LOGDIR%/%JOBNAME%'
            stat_file = open(job_name_ptrn + '_STAT', 'w')
            stat_file.write('{0:.0f}\\n'.format(time.time()))
            stat_file.close()
            ###################
            # Autosubmit Checkpoint
            ###################
            # Creates a new checkpoint file upton call based on the current numbers of calls to the function
            AS_CHECKPOINT_CALLS = 0
            def as_checkpoint():
                global AS_CHECKPOINT_CALLS
                global job_name_ptrn
                AS_CHECKPOINT_CALLS = AS_CHECKPOINT_CALLS + 1
                open(job_name_ptrn + '_CHECKPOINT_' + str(AS_CHECKPOINT_CALLS), 'w').close()                
            ###################
            # Autosubmit job
            ###################
            

            """)

    # expand tailer to use python3
    def as_tailer(self):
        return textwrap.dedent("""\

                ###################
                # Autosubmit tailer
                ###################

                stat_file = open(job_name_ptrn + '_STAT', 'a')
                stat_file.write('{0:.0f}\\n'.format(time.time()))
                stat_file.close()
                open(job_name_ptrn + '_COMPLETED', 'a').close()
                exit(0)
                """)


class StatisticsSnippetR:
    """
    Class to handle the statistics snippet of a job. It contains header and tailer for
    local and remote jobs
    """

    @staticmethod
    def as_header(scheduler_header, executable):
        if not executable:
            executable = "/usr/bin/env Rscript"
        return textwrap.dedent("""\
            #!{0}

            """).format(executable) + \
            scheduler_header + \
            textwrap.dedent("""\
            ###################
            # Autosubmit header
            ###################
            oldw <- getOption("warn")
            options( warn = -1 )
            leave = F
            langs <- c("C.utf8","C.UTF-8","C","en_GB","es_ES")
            i = 1
            e=""
            while (nchar(e) == 0 || leave)
            {
                e=Sys.setlocale("LC_ALL",langs[i])
                e
                i=i+1
                if (i > NROW(langs)) 
                {
                    leave=T
                }
            } 
            options( warn = oldw )
            job_name_ptrn = '%CURRENT_LOGDIR%/%JOBNAME%'
            fileConn<-file(paste(job_name_ptrn,"_STAT", sep = ''),"w")
            writeLines(toString(trunc(as.numeric(Sys.time()))), fileConn)
            close(fileConn)
            ###################
            # Autosubmit Checkpoint
            ###################
            # Creates a new checkpoint file upton call based on the current numbers of calls to the function
            AS_CHECKPOINT_CALLS = 0
            as_checkpoint <- function() {
                AS_CHECKPOINT_CALLS <<- AS_CHECKPOINT_CALLS + 1
                fileConn<-file(paste(job_name_ptrn,"_CHECKPOINT_",AS_CHECKPOINT_CALLS, sep = ''),"w")
                close(fileConn)
            }
            ###################
            # Autosubmit job
            ###################

            """)

    @staticmethod
    def as_tailer():
        return textwrap.dedent("""\

            ###################
            # Autosubmit tailer
            ###################

            fileConn<-file(paste(job_name_ptrn,"_STAT", sep = ''),"a")
            writeLines(toString(trunc(as.numeric(Sys.time()))), fileConn)
            close(fileConn)

            fileConn<-file(paste(job_name_ptrn,'_COMPLETED', sep = ''), 'a')
            close(fileConn)
            quit(save = 'no', status = 0)
            """)


class StatisticsSnippetEmpty:
    """
    Class to handle the statistics snippet of a job. It contains header and footer for
    local and remote jobs
    """

    @staticmethod
    def as_header(scheduler_header, executable):
        if not executable:
            executable = "/bin/bash"
        return textwrap.dedent("""\
            #!{0}

            """).format(executable) + \
            scheduler_header

    @staticmethod
    def as_tailer():
        return ''


def parse_output_number(string_number):
    """
    Parses number in format 1.0K 1.0M 1.0G

    :param string_number: String representation of number
    :type string_number: str
    :return: number in float format
    :rtype: float
    """
    number = 0.0
    if string_number:
        last_letter = string_number.strip()[-1]
        multiplier = 1
        if last_letter == "G":
            multiplier = 1000000000
            number = string_number[:-1]
        elif last_letter == "M":
            multiplier = 1000000
            number = string_number[:-1]
        elif last_letter == "K":
            multiplier = 1000
            number = string_number[:-1]
        else:
            number = string_number
        try:
            number = float(number) * multiplier
        except Exception as exp:
            number = 0.0
            pass
    return number


def increase_wallclock_by_chunk(current, increase, chunk):
    """
    Receives the wallclock times an increases it according to a quantity times the number of the current chunk.
    The result cannot be larger than 48:00. 
    If Chunk = 0 then no increment.

    :param current: WALLCLOCK HH:MM 
    :type current: str  
    :param increase: WCHUNKINC HH:MM 
    :type increase: str  
    :param chunk: chunk number 
    :type chunk: int   
    :return: HH:MM wallclock 
    :rtype: str 
    """
    # Pipeline is not testing this since mock is not well-made
    try:
        if current and increase and chunk and chunk > 0:
            wallclock = current.split(":")
            increase = increase.split(":")
            current_time = datetime.timedelta(
                hours=int(wallclock[0]), minutes=int(wallclock[1]))
            increase_time = datetime.timedelta(
                hours=int(increase[0]), minutes=int(increase[1])) * (chunk - 1)
            final_time = current_time + increase_time
            hours = int(final_time.total_seconds() // 3600)
            minutes = int((final_time.total_seconds() // 60) - (hours * 60))
            if hours > 48 or (hours >= 48 and minutes > 0):
                hours = 48
                minutes = 0
            return "%02d:%02d" % (hours, minutes)
        return current
    except Exception as exp:
        # print(exp)
        return current



#!/usr/bin/env python3

# Copyright 2015-2020 Earth Sciences Department, BSC-CNS
# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

from abc import ABCMeta, abstractmethod
import autosubmit.history.database_managers.database_models as Models
import traceback
from .internal_logging import Logging
from .database_managers.database_manager import DEFAULT_LOCAL_ROOT_DIR, DEFAULT_HISTORICAL_LOGS_DIR

class PlatformInformationHandler:
  def __init__(self, strategy):
    self._strategy = strategy
  
  @property
  def strategy(self):
    return self._strategy
  
  @strategy.setter
  def strategy(self, strategy):
    self._strategy = strategy
  
  def execute_distribution(self, job_data_dc, job_data_dcs_in_wrapper, slurm_monitor):
    return self._strategy.apply_distribution(job_data_dc, job_data_dcs_in_wrapper, slurm_monitor)
  

class Strategy(metaclass=ABCMeta):
  """ Strategy Interface """

  def __init__(self, historiclog_dir_path=DEFAULT_HISTORICAL_LOGS_DIR):
    self.historiclog_dir_path = historiclog_dir_path

  @abstractmethod
  def apply_distribution(self, job_data_dc, job_data_dcs_in_wrapper, slurm_monitor):
    pass

  def set_job_data_dc_as_processed(self, job_data_dc, original_ssh_output):
    job_data_dc.platform_output = original_ssh_output
    job_data_dc.rowstatus = Models.RowStatus.PROCESSED
    return job_data_dc
  
  def set_job_data_dc_as_process_failed(self, job_data_dc, original_ssh_output):
    job_data_dc.platform_output = original_ssh_output
    job_data_dc.rowstatus = Models.RowStatus.FAULTY
    return job_data_dc
  
  def get_calculated_weights_of_jobs_in_wrapper(self, job_data_dcs_in_wrapper):
    """ Based on computational weight: running time in seconds * number of cpus. """
    total_weight = sum(job.computational_weight for job in job_data_dcs_in_wrapper)
    if total_weight == 0:
      total_weight = 1
    return {job.job_name: round(job.computational_weight/total_weight, 4) for job in job_data_dcs_in_wrapper}


class SingleAssociationStrategy(Strategy):

  def __init__(self, historiclog_dir_path=DEFAULT_HISTORICAL_LOGS_DIR):
    super(SingleAssociationStrategy, self).__init__(historiclog_dir_path=historiclog_dir_path)

  def apply_distribution(self, job_data_dc, job_data_dcs_in_wrapper, slurm_monitor):
    try:
      if len(job_data_dcs_in_wrapper) > 0:
        return []
      # job_data_dc.submit = slurm_monitor.header.submit
      # job_data_dc.start = slurm_monitor.header.start
      # job_data_dc.finish = slurm_monitor.header.finish
      job_data_dc.ncpus = slurm_monitor.header.ncpus
      job_data_dc.nnodes = slurm_monitor.header.nnodes
      job_data_dc.energy = slurm_monitor.header.energy
      job_data_dc.MaxRSS = max(slurm_monitor.header.MaxRSS, slurm_monitor.batch.MaxRSS if slurm_monitor.batch else 0, slurm_monitor.extern.MaxRSS if slurm_monitor.extern else 0) # TODO: Improve this rule
      job_data_dc.AveRSS = max(slurm_monitor.header.AveRSS, slurm_monitor.batch.AveRSS if slurm_monitor.batch else 0, slurm_monitor.extern.AveRSS if slurm_monitor.extern else 0)        
      job_data_dc = self.set_job_data_dc_as_processed(job_data_dc, slurm_monitor.original_input)
      return [job_data_dc]
    except Exception as exp:
      Logging("strategies", self.historiclog_dir_path).log("SingleAssociationStrategy failed for {0}. Using ssh_output: {1}. Exception message: {2}".format(job_data_dc.job_name, slurm_monitor.original_input, str(exp)), 
                              traceback.format_exc())
      job_data_dc = self.set_job_data_dc_as_process_failed(job_data_dc, slurm_monitor.original_input)
      return [job_data_dc]

class StraightWrapperAssociationStrategy(Strategy):

  def __init__(self, historiclog_dir_path=DEFAULT_HISTORICAL_LOGS_DIR):
    super(StraightWrapperAssociationStrategy, self).__init__(historiclog_dir_path=historiclog_dir_path)

  def apply_distribution(self, job_data_dc, job_data_dcs_in_wrapper, slurm_monitor):
    """ """
    try:
      if len(job_data_dcs_in_wrapper) != slurm_monitor.step_count:
        return []
      result = []
      computational_weights = self.get_calculated_weights_of_jobs_in_wrapper(job_data_dcs_in_wrapper)      
      for job_dc, step in zip(job_data_dcs_in_wrapper, slurm_monitor.steps):        
        job_dc.energy = step.energy + computational_weights.get(job_dc.job_name, 0) * slurm_monitor.extern.energy
        job_dc.AveRSS = step.AveRSS
        job_dc.MaxRSS = step.MaxRSS        
        job_dc.platform_output = ""
        if job_dc.job_name == job_data_dc.job_name:
          job_data_dc.energy = job_dc.energy
          job_data_dc.AveRSS = job_dc.AveRSS
          job_data_dc.MaxRSS = job_dc.MaxRSS        
        result.append(job_dc)
      job_data_dc = self.set_job_data_dc_as_processed(job_data_dc, slurm_monitor.original_input)
      result.append(job_data_dc)
      return result
    except Exception as exp:
      Logging("strategies", self.historiclog_dir_path).log("StraightWrapperAssociationStrategy failed for {0}. Using ssh_output: {1}. Exception message: {2}".format(job_data_dc.job_name, 
                                    slurm_monitor.original_input, 
                                    str(exp)), 
                              traceback.format_exc())
      job_data_dc = self.set_job_data_dc_as_process_failed(job_data_dc, slurm_monitor.original_input)
      return [job_data_dc]

class GeneralizedWrapperDistributionStrategy(Strategy):

  def __init__(self, historiclog_dir_path=DEFAULT_HISTORICAL_LOGS_DIR):
    super(GeneralizedWrapperDistributionStrategy, self).__init__(historiclog_dir_path=historiclog_dir_path)

  def apply_distribution(self, job_data_dc, job_data_dcs_in_wrapper, slurm_monitor):
    try:
      result = []
      computational_weights = self.get_calculated_weights_of_jobs_in_wrapper(job_data_dcs_in_wrapper)
      for job_dc in job_data_dcs_in_wrapper:
        job_dc.energy = round(computational_weights.get(job_dc.job_name, 0) * slurm_monitor.total_energy,2)
        job_dc.platform_output = "" 
        if job_dc.job_name == job_data_dc.job_name:
          job_data_dc.energy = job_dc.energy      
        result.append(job_dc)
      job_data_dc = self.set_job_data_dc_as_processed(job_data_dc, slurm_monitor.original_input)
      result.append(job_data_dc)
      return result
    except Exception as exp:
      Logging("strategies", self.historiclog_dir_path).log("GeneralizedWrapperDistributionStrategy failed for {0}. Using ssh_output: {1}. Exception message: {2}".format(job_data_dc.job_name, slurm_monitor.original_input, str(exp)), 
                              traceback.format_exc())
      job_data_dc = self.set_job_data_dc_as_process_failed(job_data_dc, slurm_monitor.original_input)
      return [job_data_dc]

class TwoDimWrapperDistributionStrategy(Strategy):

  def __init__(self, historiclog_dir_path=DEFAULT_HISTORICAL_LOGS_DIR):
    super(TwoDimWrapperDistributionStrategy, self).__init__(historiclog_dir_path=historiclog_dir_path)
    self.jobs_per_level = None

  def apply_distribution(self, job_data_dc, job_data_dcs_in_wrapper, slurm_monitor):
    try:
      result = []        
      self.jobs_per_level = self.get_jobs_per_level(job_data_dcs_in_wrapper)
      if len(self.jobs_per_level) != slurm_monitor.step_count:
        return []
      comp_weight_per_level = self.get_comp_weight_per_level(self.jobs_per_level)
      level_energy = []
      for i, step in enumerate(slurm_monitor.steps):
        level_energy.append(step.energy + comp_weight_per_level[i] * slurm_monitor.extern.energy)
      for i, jobs in enumerate(self.jobs_per_level):
        weights = self.get_comp_weight_per_group_of_job_dcs(jobs)
        for j, job_dc in enumerate(jobs):
          job_dc.energy = round(level_energy[i] * weights[j], 2)
          if job_dc.job_name == job_data_dc.job_name:
            job_data_dc.energy = job_dc.energy    
          result.append(job_dc)   
      job_data_dc = self.set_job_data_dc_as_processed(job_data_dc, slurm_monitor.original_input)
      result.append(job_data_dc)
      return result
    except Exception as exp:
      Logging("strategies", self.historiclog_dir_path).log("TwoDimWrapperDistributionStrategy failed for {0}. Using ssh_output: {1}. Exception message: {2}".format(job_data_dc.job_name, slurm_monitor.original_input, str(exp)), 
                              traceback.format_exc())
      job_data_dc = self.set_job_data_dc_as_process_failed(job_data_dc, slurm_monitor.original_input)
      return [job_data_dc]

  def get_jobs_per_level(self, job_data_dcs_in_wrapper):
    """ List of Lists, index of list is the level. """
    job_name_to_object = {job.job_name: job for job in job_data_dcs_in_wrapper}         
    levels = []    
    roots_dcs = self._get_roots(job_data_dcs_in_wrapper)    
    levels.append(roots_dcs)
    next_level = self.get_level(roots_dcs, job_name_to_object)      
    while len(next_level) > 0:      
      levels.append([job for job in next_level])      
      next_level = self.get_level(next_level, job_name_to_object)            
    return levels
  
  def _get_roots(self, job_data_dcs_in_wrapper):
    children_names = self._get_all_children(job_data_dcs_in_wrapper)
    return [job for job in job_data_dcs_in_wrapper if job.job_name not in children_names]

  def _get_all_children(self, job_data_dcs_in_wrapper):
    result = []
    for job_dc in job_data_dcs_in_wrapper:
      result.extend(job_dc.children_list)
    return result
  
  def get_comp_weight_per_group_of_job_dcs(self, jobs):
    total = sum(job.computational_weight for job in jobs)
    return [round(job.computational_weight/total, 4) for job in jobs]

  def get_comp_weight_per_level(self, jobs_per_level):
    level_weight = []
    total_weight = 0
    for jobs in jobs_per_level:
      computational_weight = sum(job.computational_weight for job in jobs)
      total_weight += computational_weight
      level_weight.append(computational_weight)    
    return [round(weight/total_weight, 4) for weight in level_weight]

  def get_level(self, previous_level_dcs, job_name_to_object):
    children_names = []        
    for job_dc in previous_level_dcs:
      children_names.extend(job_dc.children_list)        
    level_dcs = [job_name_to_object[job_name] for job_name in children_names if job_name in job_name_to_object]    
    return level_dcs

    


import os
import stat

_fd_types = (
    ('REG', stat.S_ISREG),
    ('FIFO', stat.S_ISFIFO),
    ('DIR', stat.S_ISDIR),
    ('CHR', stat.S_ISCHR),
    ('BLK', stat.S_ISBLK),
    ('LNK', stat.S_ISLNK),
    ('SOCK', stat.S_ISSOCK)
)

def fd_table_status():
    result = []
    for fd in range(1024):
        try:
            s = os.fstat(fd)
        except:
            continue
        for fd_type, func in _fd_types:
            if func(s.st_mode):
                break
        else:
            fd_type = str(s.st_mode)
        result.append((fd, fd_type))
    return result

def fd_table_status_logify(fd_table_result):
    return ('Open file handles: ' +
            ', '.join(['{0}: {1}'.format(*i) for i in fd_table_result]))

def fd_table_status_str():
    return fd_table_status_logify(fd_table_status())

if __name__=='__main__':
    print(fd_table_status_str())


#!/usr/bin/env python3

# Copyright 2017-2020 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.
import locale
import os
from contextlib import suppress
from time import sleep
from typing import List, Union

from autosubmit.job.job_common import Status
from autosubmit.platforms.paramiko_platform import ParamikoPlatform
from autosubmit.platforms.headers.pjm_header import PJMHeader
from autosubmit.platforms.wrappers.wrapper_factory import PJMWrapperFactory
from log.log import AutosubmitCritical, AutosubmitError, Log

class PJMPlatform(ParamikoPlatform):
    """
    Class to manage jobs to host using PJM scheduler

    :param expid: experiment's identifier
    :type expid: str
    """


    def __init__(self, expid, name, config):
        ParamikoPlatform.__init__(self, expid, name, config)
        self.mkdir_cmd = None
        self.get_cmd = None
        self.put_cmd = None
        self._submit_hold_cmd = None
        self._submit_command_name = None
        self._submit_cmd = None
        self._checkhost_cmd = None
        self.cancel_cmd = None
        self._header = PJMHeader()
        self._wrapper = PJMWrapperFactory(self)
        #https://software.fujitsu.com/jp/manual/manualfiles/m220008/j2ul2452/02enz007/j2ul-2452-02enz0.pdf page 16
        self.job_status = dict()
        self.job_status['COMPLETED'] = ['EXT']
        self.job_status['RUNNING'] = ['RNO','RNE','RUN']
        self.job_status['QUEUING'] = ['ACC','QUE', 'RNA', 'RNP','HLD'] # TODO NOT SURE ABOUT HOLD HLD
        self.job_status['FAILED'] = ['ERR','CCL','RJT']
        self._pathdir = "\$HOME/LOG_" + self.expid
        self._allow_arrays = False
        self._allow_wrappers = True # NOT SURE IF WE NEED WRAPPERS
        self.update_cmds()
        self.config = config
        exp_id_path = os.path.join(self.config.get("LOCAL_ROOT_DIR"), self.expid)
        tmp_path = os.path.join(exp_id_path, "tmp")
        self._submit_script_path = os.path.join(
            tmp_path, self.config.get("LOCAL_ASLOG_DIR"), "submit_" + self.name + ".sh")
        self._submit_script_file = open(self._submit_script_path, 'wb').close()

    def submit_error(self,output):
        """
        Check if the output of the submit command contains an error message.
        :param output: output of the submit cmd
        :return: boolean
        """
        return not all(part.lower() in output.lower() for part in ["pjsub", "[INFO] PJM 0000"])



    def process_batch_ready_jobs(self,valid_packages_to_submit,failed_packages,error_message="",hold=False):
        """
        Retrieve multiple jobs identifiers.
        :param valid_packages_to_submit:
        :param failed_packages:
        :param error_message:
        :param hold:
        :return:
        """
        try:
            valid_packages_to_submit = [ package for package in valid_packages_to_submit ]
            if len(valid_packages_to_submit) > 0:
                try:
                    jobs_id = self.submit_Script(hold=hold)
                except AutosubmitError as e:
                    jobnames = [job.name for job in valid_packages_to_submit[0].jobs]
                    for jobname in jobnames:
                        jobid = self.get_jobid_by_jobname(jobname)
                        #cancel bad submitted job if jobid is encountered
                        for id_ in jobid:
                            self.cancel_job(id_)
                    jobs_id = None
                    self.connected = False
                    if e.trace is not None:
                        has_trace_bad_parameters = self.submit_error(e.trace)
                    else:
                        has_trace_bad_parameters = False
                    if e.message is not None:
                        has_message_bad_parameters = self.submit_error(e.message)
                    else:
                        has_message_bad_parameters = False
                    if has_trace_bad_parameters or has_message_bad_parameters or e.message.lower().find("invalid partition") != -1 or e.message.lower().find(" invalid qos") != -1 or e.message.lower().find("scheduler is not installed") != -1 or e.message.lower().find("failed") != -1 or e.message.lower().find("not available") != -1:
                        error_msg = ""
                        for package_tmp in valid_packages_to_submit:
                            for job_tmp in package_tmp.jobs:
                                if job_tmp.section not in error_msg:
                                    error_msg += job_tmp.section + "&"
                        if has_trace_bad_parameters:
                            error_message+="Check job and queue specified in jobs.conf. Sections that could be affected: {0}".format(error_msg[:-1])
                        else:
                            error_message+="\ncheck that {1} platform has set the correct scheduler. Sections that could be affected: {0}".format(
                                    error_msg[:-1], self.name)

                        if e.trace is None:
                            e.trace = ""
                        raise AutosubmitCritical(error_message,7014,e.message+"\n"+str(e.trace))
                except IOError as e:
                    raise AutosubmitError(
                        "IO issues ", 6016, str(e))
                except BaseException as e:
                    if str(e).find("scheduler") != -1:
                        raise AutosubmitCritical("Are you sure that [{0}] scheduler is the correct type for platform [{1}]?.\n Please, double check that {0} is loaded for {1} before autosubmit launch any job.".format(self.type.upper(),self.name.upper()),7070)
                    raise AutosubmitError(
                        "Submission failed, this can be due a failure on the platform", 6015, str(e))
                if jobs_id is None or len(jobs_id) <= 0:
                    raise AutosubmitError(
                        "Submission failed, this can be due a failure on the platform", 6015,"Jobs_id {0}".format(jobs_id))
                i = 0
                if hold:
                    sleep(10)

                for package in valid_packages_to_submit:
                    for job in package.jobs:
                        job.hold = hold
                        job.id = str(jobs_id[i])
                        job.status = Status.SUBMITTED
                        job.write_submit_time(hold=hold)
                    i += 1
            save = True
        except AutosubmitError as e:
            raise
        except AutosubmitCritical as e:
            raise
        except Exception as e:
            raise AutosubmitError("{0} submission failed".format(self.name), 6015, str(e))
        return save,valid_packages_to_submit

    def open_submit_script(self):
        self._submit_script_file = open(self._submit_script_path, 'wb').close()
        self._submit_script_file = open(self._submit_script_path, 'ab')

    def get_submit_script(self):
        self._submit_script_file.close()
        os.chmod(self._submit_script_path, 0o750)
        return os.path.join(self.config.get("LOCAL_ASLOG_DIR"), os.path.basename(self._submit_script_path))

    def submit_job(self, job, script_name, hold=False, export="none"):
        """
        Submit a job from a given job object.

        :param export:
        :param job: job object
        :type job: autosubmit.job.job.Job
        :param script_name: job script's name
        :rtype scriptname: str
        :param hold: send job hold
        :type hold: boolean
        :return: job id for the submitted job
        :rtype: int
        """
        self.get_submit_cmd(script_name, job, hold=hold, export=export)
        return None


    def submit_Script(self, hold=False):
        # type: (bool) -> Union[List[str], str]
        """
        Sends a Submit file Script, execute it  in the platform and retrieves the Jobs_ID of all jobs at once.

        :param hold: if True, the job will be held
        :type hold: bool
        :return: job id for  submitted jobs
        :rtype: list(str)
        """
        try:
            self.send_file(self.get_submit_script(), False)
            cmd = os.path.join(self.get_files_path(),
                               os.path.basename(self._submit_script_path))
            try:
                self.send_command(cmd)
            except AutosubmitError as e:
                raise
            except AutosubmitCritical as e:
                raise
            except Exception as e:
                raise
            jobs_id = self.get_submitted_job_id(self.get_ssh_output())
            return jobs_id
        except IOError as e:
            raise AutosubmitError("Submit script is not found, retry again in next AS iteration", 6008, str(e))
        except AutosubmitError as e:
            raise
        except AutosubmitCritical as e:
            raise
        except Exception as e:
            raise AutosubmitError("Submit script is not found, retry again in next AS iteration", 6008, str(e))
    def check_remote_log_dir(self):
        """
        Creates log dir on remote host
        """

        try:
            # Test if remote_path exists
            self._ftpChannel.chdir(self.remote_log_dir)
        except IOError as e:
            try:
                if self.send_command(self.get_mkdir_cmd()):
                    Log.debug('{0} has been created on {1} .',
                              self.remote_log_dir, self.host)
                else:
                    raise AutosubmitError("SFTP session not active ", 6007, "Could not create the DIR {0} on HPC {1}'.format(self.remote_log_dir, self.host)".format(
                        self.remote_log_dir, self.host))
            except BaseException as e:
                raise AutosubmitError(
                    "SFTP session not active ", 6007, str(e))

    def update_cmds(self):
        """
        Updates commands for platforms
        """
        self.root_dir = os.path.join(
            self.scratch, self.project_dir, self.user, self.expid)
        self.remote_log_dir = os.path.join(self.root_dir, "LOG_" + self.expid)
        self.cancel_cmd = "pjdel "
        self._checkhost_cmd = "echo 1"
        self._submit_cmd = 'cd {0} ; pjsub  '.format(self.remote_log_dir)
        self._submit_command_name = "pjsub"
        self._submit_hold_cmd = 'cd {0} ; pjsub  '.format(self.remote_log_dir)
        self.put_cmd = "scp"
        self.get_cmd = "scp"
        self.mkdir_cmd = "mkdir -p " + self.remote_log_dir

    def hold_job(self, job):
        try:
            cmd = "pjrls {0} ; sleep 2 ; pjhold -R ASHOLD {0}".format(job.id)
            self.send_command(cmd)
            job_status = self.check_job(job, submit_hold_check=True)
            if job_status == Status.RUNNING:
                self.send_command("{0} {1}".format(self.cancel_cmd,job.id))
                return False
            elif job_status == Status.FAILED:
                return False
            cmd = self.get_queue_status_cmd(job.id)
            self.send_command(cmd)
        except BaseException as e:
            try:
                self.send_command("{0} {1}".format(self.cancel_cmd,job.id))
                raise AutosubmitError(
                    "Can't hold jobid:{0}, canceling job".format(job.id), 6000, str(e))
            except BaseException as e:
                raise AutosubmitError(
                    "Can't cancel the jobid: {0}".format(job.id), 6000, str(e))
            except AutosubmitError as e:
                raise

    def get_checkhost_cmd(self):
        return self._checkhost_cmd

    def get_mkdir_cmd(self):
        return self.mkdir_cmd

    def get_remote_log_dir(self):
        return self.remote_log_dir

    def parse_job_output(self, output):
        return output.strip().split()[0].strip()

    def parse_job_finish_data(self, output, packed):
        return 0, 0, 0, 0, 0, 0, dict(), False

    def queuing_reason_cancel(self, reason):
        try:
            if len(reason.split('(', 1)) > 1:
                reason = reason.split('(', 1)[1].split(')')[0]
                if 'Invalid' in reason or reason in ['ANOTHER JOB STARTED','DELAY','DEADLINE SCHEDULE STARTED','ELAPSE LIMIT EXCEEDED','FILE IO ERROR','GATE CHECK','IMPOSSIBLE SCHED','INSUFF CPU','INSUFF MEMORY','INSUFF NODE','INSUFF','INTERNAL ERROR','INVALID HOSTFILE','LIMIT OVER MEMORY','LOST COMM','NO CURRENT DIR','NOT EXIST','RSCGRP NOT EXIST','RSCGRP STOP','RSCUNIT','USER','EXCEED','WAIT SCHED']:
                    return True
            return False
        except Exception as e:
            return False
    def get_queue_status(self, in_queue_jobs, list_queue_jobid, as_conf):
        if not in_queue_jobs:
            return
        cmd = self.get_queue_status_cmd(list_queue_jobid)
        self.send_command(cmd)
        queue_status = self._ssh_output
        for job in in_queue_jobs:
            reason = self.parse_queue_reason(queue_status, job.id)
            if job.queuing_reason_cancel(reason):
                Log.printlog(f"Job {job.name} will be cancelled and set to FAILED as it was queuing due to {reason}",6000)
                self.send_command(self.cancel_cmd + " {0}".format(job.id))
                job.new_status = Status.FAILED
                job.update_status(as_conf)
            elif reason.find('ASHOLD') != -1:
                job.new_status = Status.HELD
                if not job.hold:
                    self.send_command("{0} {1}".format(self.cancel_cmd,job.id))
                    job.new_status = Status.QUEUING  # If it was HELD and was released, it should be QUEUING next.
    def parse_Alljobs_output(self, output, job_id):
        status = ""
        try:
            status = [x.split()[1] for x in output.splitlines()
                      if x.split()[0] == str(job_id)]
        except BaseException as e:
            pass
        if len(status) == 0:
            return status
        return status[0]

    def parse_joblist(self, job_list):
        """
        Convert a list of job_list to job_list_cmd
        :param job_list: list of jobs
        :type job_list: list
        :return: job status
        :rtype: str
        """
        job_list_cmd = ""
        for job, job_prev_status in job_list:
            if job.id is None:
                job_str = "0"
            else:
                job_str = str(job.id)
            job_list_cmd += job_str + "+"
        if job_list_cmd[-1] == "+":
            job_list_cmd = job_list_cmd[:-1]

        return job_list_cmd
    def _check_jobid_in_queue(self, ssh_output, job_list_cmd):
        for job in job_list_cmd.split('+'):
            if job not in ssh_output:
                return False
        return True

    def get_submitted_job_id(self, outputlines):
        try:
            jobs_id = []
            for output in outputlines.splitlines():
                if not self.submit_error(output):
                    jobs_id.append(int(output.split()[5]))

            return jobs_id
        except IndexError:
            raise AutosubmitCritical(
                "Submission failed. There are issues on your config file", 7014)

    def get_submit_cmd(self, job_script, job, hold=False, export=""):
        if (export is None or export.lower() == "none") or len(export) == 0:
            export = ""
        else:
            export += " ; "


        with suppress(BaseException):
            lang = locale.getlocale()[1]
            if lang is None:
                lang = locale.getdefaultlocale()[1]
                if lang is None:
                    lang = 'UTF-8'
            if not hold:
                self._submit_script_file.write((export + self._submit_cmd + job_script + "\n").encode(lang))
            else:
                self._submit_script_file.write((export + self._submit_hold_cmd + job_script + "\n").encode(lang))


    def get_checkAlljobs_cmd(self, jobs_id):
        # jobs_id = "jobid1+jobid2+jobid3"
        # -H == sacct
        return "pjstat -H -v --choose jid,st,ermsg --filter \"jid={0}\" > as_checkalljobs.txt ; pjstat -v --choose jid,st,ermsg --filter \"jid={0}\" >> as_checkalljobs.txt ; cat as_checkalljobs.txt ; rm as_checkalljobs.txt".format(jobs_id)

    def get_queue_status_cmd(self, job_id):
        return self.get_checkAlljobs_cmd(job_id)

    def get_jobid_by_jobname_cmd(self, job_name):
        return 'pjstat -v --choose jid,st,ermsg --filter \"jnam={0}\"'.format(job_name)


    def cancel_job(self, job_id):
        return '{0} {1}'.format(self.cancel_cmd,job_id)

    #def get_job_energy_cmd(self, job_id):
    #    return 'sacct -n --jobs {0} -o JobId%25,State,NCPUS,NNodes,Submit,Start,End,ConsumedEnergy,MaxRSS%25,AveRSS%25'.format(job_id)

    def parse_queue_reason(self, output, job_id):
        # split() is used to remove the trailing whitespace but also \t and multiple spaces
        # split(" ") is not enough
        reason = [x.split()[2] for x in output.splitlines()
                  if x.split()[0] == str(job_id)]
        # In case of duplicates.. we take the first one
        if len(reason) > 0:
            return reason[0]
        return reason

    def wrapper_header(self,**kwargs):
        if method == 'srun':
            language = "#!/bin/bash"
            return \
                language + """
###############################################################################
#              {0}
###############################################################################
#
#PJM -N {0}
{1}
{8}
#PJM -g {2}
#PJM -o {0}.out
#PJM -e {0}.err
#PJM -elapse {3}:00
#PJM --mpi "proc=%NUMPROC%"
#PJM --mpi "max-proc-per-node={7}"
{5}
{6}

#
###############################################################################
                """.format(filename, queue, project, wallclock, num_procs, dependency,
                           '\n'.ljust(13).join(str(s) for s in directives), threads,partition)
        else:
            language = "#!/usr/bin/env python3"
            return \
                language + """
###############################################################################
#              {0}
###############################################################################
#
#PJM -N {0}
{1}
{8}
#PJM -g {2}
#PJM -o {0}.out
#PJM -e {0}.err
#PJM -elapse {3}:00
#PJM --mpi "proc=%NUMPROC%"
#PJM --mpi "max-proc-per-node={7}"
{5}
{6}
#
###############################################################################
            """.format(filename, queue, project, wallclock, num_procs, dependency,
                       '\n'.ljust(13).join(str(s) for s in directives), threads,partition)

    @staticmethod
    def allocated_nodes():
        return """os.system("scontrol show hostnames $SLURM_JOB_NODELIST > node_list_{0}".format(node_id))"""

    def check_file_exists(self, filename,wrapper_failed=False):
        file_exist = False
        sleeptime = 5
        retries = 0
        max_retries = 3
        while not file_exist and retries < max_retries:
            try:
                # This return IOError if path doesn't exist
                self._ftpChannel.stat(os.path.join(
                    self.get_files_path(), filename))
                file_exist = True
            except IOError as e:  # File doesn't exist, retry in sleeptime
                Log.debug("{2} File does not exist.. waiting {0}s for a new retry (retries left: {1})", sleeptime,
                          max_retries - retries, os.path.join(self.get_files_path(), filename))
                if not wrapper_failed:
                    sleep(sleeptime)
                    sleeptime = sleeptime + 5
                    retries = retries + 1
                else:
                    retries = 9999
            except BaseException as e:  # Unrecoverable error
                if str(e).lower().find("garbage") != -1:
                    if not wrapper_failed:
                        sleep(sleeptime)
                        sleeptime = sleeptime + 5
                        retries = retries + 1
                else:
                    Log.printlog("remote logs {0} couldn't be recovered".format(filename), 6001)
                    file_exist = False  # won't exist
                    retries = 999  # no more retries
        return file_exist



import functools
import inspect
from collections import defaultdict
from typing import Dict

PARAMETERS = defaultdict(defaultdict)
"""Global default dictionary holding a multi-level dictionary with the Autosubmit
parameters. At the first level we have the parameter groups.

  - ``JOB``

  - ``PLATFORM``
  
  - ``PROJECT``
  
Each entry in the ``PARAMETERS`` dictionary holds another default dictionary. Finally,
the lower level in the dictionary has a ``key=value`` where ``key`` is the parameter
name, and ``value`` the parameter documentation.

These values are used to create the Sphinx documentation for variables, as well as
to populate the comments in the Autosubmit YAML configuration files.
"""


def autosubmit_parameters(cls=None, *, parameters: Dict):
    """Decorator for Autosubmit configuration parameters defined in a class.

    This is useful for parameters that are not defined in a single function or
    class (e.g. parameters that are created on-the-fly in functions)."""

    def wrap(cls):
        parameters = wrap.parameters

        for group, group_parameters in parameters.items():
            group = group.upper()

            if group not in PARAMETERS:
                PARAMETERS[group] = defaultdict(defaultdict)

            for parameter_name, parameter_value in group_parameters.items():
                if parameter_name not in PARAMETERS[group]:
                    PARAMETERS[group][parameter_name] = parameter_value.strip()

        return cls

    wrap.parameters = parameters

    # NOTE: This is not reachable code, as the parameters must be defined!
    # if cls is not None:
    #     raise ValueError(f'You must provide a list of parameters')

    return wrap


def autosubmit_parameter(func=None, *, name, group=None):
    """Decorator for Autosubmit configuration parameters.

    Used to annotate properties of classes

    Attributes:
        func (Callable): wrapped function. Always ``None`` due to how we call the decorator.
        name (Union[str, List[str]]): parameter name.
        group (str): group name. Default to caller module name.
    """
    if group is None:
        stack = inspect.stack()
        group: str = stack[1][0].f_locals['__qualname__'].rsplit('.', 1)[-1]

    group = group.upper()

    if group not in PARAMETERS:
        PARAMETERS[group] = defaultdict(defaultdict)

    names = name
    if type(name) is not list:
        names = [name]

    for parameter_name in names:
        if parameter_name not in PARAMETERS[group]:
            PARAMETERS[group][parameter_name] = None

    def parameter_decorator(wrapped_func):
        parameter_group = parameter_decorator.__group
        parameter_names = parameter_decorator.__names
        for p_name in parameter_names:
            if wrapped_func.__doc__:
                PARAMETERS[parameter_group][p_name] = wrapped_func.__doc__.strip().split('\n')[0]

        # Delete the members created as we are not using them hereafter
        del parameter_decorator.__group
        del parameter_decorator.__names

        @functools.wraps(wrapped_func)
        def wrapper(*args, **kwargs):
            return wrapped_func(*args, **kwargs)

        return wrapper

    parameter_decorator.__group = group
    parameter_decorator.__names = names

    return parameter_decorator



import os
from unittest import TestCase


#class TestSetup(TestCase):
    #def test_setup_check_works(self):
    #    exit_code = run_setup_check()
    #    self.assertEqual(0, exit_code)


def run_setup_check():
    return os.system('python ' + get_directory_of_this_file() + '/../../setup.py check')


def get_directory_of_this_file():
    return os.path.dirname(os.path.realpath(__file__))



#!/usr/bin/env python3

# Copyright 2015-2020 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

from autosubmit.platforms.wrappers.wrapper_builder import WrapperDirector, PythonVerticalWrapperBuilder, \
    PythonHorizontalWrapperBuilder, PythonHorizontalVerticalWrapperBuilder, PythonVerticalHorizontalWrapperBuilder, \
    BashHorizontalWrapperBuilder, BashVerticalWrapperBuilder, SrunHorizontalWrapperBuilder,SrunVerticalHorizontalWrapperBuilder


class WrapperFactory(object):

    def __init__(self, platform):
        self.as_conf = None
        self.platform = platform
        self.wrapper_director = WrapperDirector()
        self.exception = "This type of wrapper is not supported for this platform"

    def get_wrapper(self, wrapper_builder, **kwargs):
        wrapper_data = kwargs['wrapper_data']
        kwargs['allocated_nodes'] = self.allocated_nodes()
        kwargs['dependency'] = self.dependency(kwargs['dependency'])
        kwargs['queue'] = self.queue(kwargs['queue'])
        kwargs['partition'] = self.partition(wrapper_data['PARTITION'])
        kwargs["exclusive"] = self.exclusive(wrapper_data['EXCLUSIVE'])
        kwargs["custom_directives"] = self.custom_directives(wrapper_data["CUSTOM_DIRECTIVES"])
        kwargs["executable"] = wrapper_data["EXECUTABLE"]
        kwargs['nodes'] = self.nodes(wrapper_data['NODES'])
        kwargs['tasks'] = self.tasks(wrapper_data['TASKS'])
        kwargs['threads'] = self.threads(kwargs['threads'])
        if str(kwargs['num_processors']).isdigit():
            kwargs['num_processors_value'] = int(kwargs['num_processors'])
        else:
            kwargs['num_processors_value'] = 1
        if str(wrapper_data['NODES']).isdigit() and int(wrapper_data['NODES']) > 1 and kwargs['num_processors'] == '1':
            kwargs['num_processors'] = "#"
        else:
            kwargs['num_processors'] = self.processors(kwargs['num_processors'])
        kwargs['header_directive'] = self.header_directives(**kwargs)
        builder = wrapper_builder(**kwargs)
        return self.wrapper_director.construct(builder)

    def vertical_wrapper(self, **kwargs):
        raise NotImplemented(self.exception)

    def horizontal_wrapper(self, **kwargs):
        raise NotImplemented(self.exception)

    def hybrid_wrapper_horizontal_vertical(self, **kwargs):
        raise NotImplemented(self.exception)

    def hybrid_wrapper_vertical_horizontal(self, **kwargs):
        raise NotImplemented(self.exception)

    def header_directives(self, **kwargs):
        pass

    def allocated_nodes(self):
        return ''

    def dependency(self, dependency):
        return '#' if dependency is None else self.dependency_directive(dependency)
    def queue(self, queue):
        return '#' if not queue else self.queue_directive(queue)
    def processors(self, processors):
        return '#' if not processors or processors == "0" else self.processors_directive(processors)
    def nodes(self, nodes):
        return '#' if not nodes else self.nodes_directive(nodes)
    def tasks(self, tasks):
        return '#' if not tasks else self.tasks_directive(tasks)
    def partition(self, partition):
        return '#' if not partition else self.partition_directive(partition)
    def threads(self, threads):
        return '#' if not threads or threads in ["0","1"] else self.threads_directive(threads)
    def exclusive(self, exclusive):
        return '#' if not exclusive or str(exclusive).lower() == "false" else self.exclusive_directive(exclusive)
    def custom_directives(self, custom_directives):
        return '#' if not custom_directives else self.get_custom_directives(custom_directives)
    def get_custom_directives(self, custom_directives):
        """
        Returns custom directives for the specified job
        :param job: Job object
        :return: String with custom directives
        """
        # There is no custom directives, so directive is empty
        if custom_directives != '':
            return '\n'.join(str(s) for s in custom_directives)
        return ""

    def dependency_directive(self, dependency):
        raise NotImplemented(self.exception)
    def queue_directive(self, queue):
        raise NotImplemented(self.exception)
    def processors_directive(self, processors):
        raise NotImplemented(self.exception)
    def nodes_directive(self, nodes):
        raise NotImplemented(self.exception)
    def tasks_directive(self, tasks):
        raise NotImplemented(self.exception)
    def partition_directive(self, partition):
        raise NotImplemented(self.exception)
    def exclusive_directive(self, exclusive):
        raise NotImplemented(self.exception)
    def threads_directive(self, threads):
        raise NotImplemented(self.exception)

class SlurmWrapperFactory(WrapperFactory):

    def vertical_wrapper(self, **kwargs):
        return PythonVerticalWrapperBuilder(**kwargs)

    def horizontal_wrapper(self, **kwargs):

        if kwargs["method"] == 'srun':
            return SrunHorizontalWrapperBuilder(**kwargs)
        else:
            return PythonHorizontalWrapperBuilder(**kwargs)

    def hybrid_wrapper_horizontal_vertical(self, **kwargs):
        return PythonHorizontalVerticalWrapperBuilder(**kwargs)

    def hybrid_wrapper_vertical_horizontal(self, **kwargs):
        if kwargs["method"] == 'srun':
            return SrunVerticalHorizontalWrapperBuilder(**kwargs)
        else:
            return PythonVerticalHorizontalWrapperBuilder(**kwargs)

    def header_directives(self, **kwargs):
        return self.platform.wrapper_header(**kwargs)

    def allocated_nodes(self):
        return self.platform.allocated_nodes()

    def dependency_directive(self, dependency):
        return '#SBATCH --dependency=afterok:{0}'.format(dependency)

    def queue_directive(self, queue):
        return '#SBATCH --qos={0}'.format(queue)
    def partition_directive(self, partition):
        return '#SBATCH --partition={0}'.format(partition)
    def exclusive_directive(self, exclusive):
        return '#SBATCH --exclusive'
    def tasks_directive(self, tasks):
        return '#SBATCH --ntasks-per-node={0}'.format(tasks)
    def nodes_directive(self, nodes):
        return '#SBATCH -N {0}'.format(nodes)
    def processors_directive(self, processors):
        return '#SBATCH -n {0}'.format(processors)
    def threads_directive(self, threads):
        return '#SBATCH --cpus-per-task={0}'.format(threads)



class LSFWrapperFactory(WrapperFactory):

    def vertical_wrapper(self, **kwargs):
        return PythonVerticalWrapperBuilder(**kwargs)

    def horizontal_wrapper(self, **kwargs):
        return PythonHorizontalWrapperBuilder(**kwargs)

    def header_directives(self, **kwargs):
        return self.platform.wrapper_header(**kwargs)

    def queue_directive(self, queue):
        return queue

    def dependency_directive(self, dependency):
        return '#BSUB -w \'done({0})\' [-ti]'.format(dependency)


class EcWrapperFactory(WrapperFactory):

    def vertical_wrapper(self, **kwargs):
        return BashVerticalWrapperBuilder(**kwargs)

    def horizontal_wrapper(self, **kwargs):
        return BashHorizontalWrapperBuilder(**kwargs)

    def header_directives(self, **kwargs):
        return self.platform.wrapper_header(**kwargs)

    def queue_directive(self, queue):
        return queue

    def dependency_directive(self, dependency):
        return '#PBS -v depend=afterok:{0}'.format(dependency)

class PJMWrapperFactory(WrapperFactory):

    def vertical_wrapper(self, **kwargs):
        return PythonVerticalWrapperBuilder(**kwargs)

    def horizontal_wrapper(self, **kwargs):

        if kwargs["method"] == 'srun':
            return SrunHorizontalWrapperBuilder(**kwargs)
        else:
            return PythonHorizontalWrapperBuilder(**kwargs)

    def hybrid_wrapper_horizontal_vertical(self, **kwargs):
        return PythonHorizontalVerticalWrapperBuilder(**kwargs)

    def hybrid_wrapper_vertical_horizontal(self, **kwargs):
        if kwargs["method"] == 'srun':
            return SrunVerticalHorizontalWrapperBuilder(**kwargs)
        else:
            return PythonVerticalHorizontalWrapperBuilder(**kwargs)

    def header_directives(self, **kwargs):
        return self.platform.wrapper_header(**kwargs)

    def allocated_nodes(self):
        return self.platform.allocated_nodes()

    #def dependency_directive(self, dependency):
    #    # There is no option for afterok in the PJM scheduler, but I think it is not needed.
    #    return '#PJM --dependency=afterok:{0}'.format(dependency)

    def queue_directive(self, queue):
        return '#PJM -L rscgrp={0}'.format(queue)

    def partition_directive(self, partition):
        return '#PJM -g {0}'.format(partition)



#!/usr/bin/env python3

# Copyright 2015-2020 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

import smtplib
import email.utils
from email.mime.text import MIMEText
from log.log import Log

class MailNotifier:
    def __init__(self, basic_config):
        self.config = basic_config

    def notify_experiment_status(self, exp_id,mail_to,platform):
        message_text = self._generate_message_experiment_status(exp_id, platform)
        message = MIMEText(message_text)
        message['From'] = email.utils.formataddr(('Autosubmit', self.config.MAIL_FROM))
        message['Subject'] = f'[Autosubmit] Warning a remote platform is malfunctioning'
        message['Date'] = email.utils.formatdate(localtime=True)
        for mail in mail_to:
            message['To'] = email.utils.formataddr((mail, mail))
            try:
                self._send_mail(self.config.MAIL_FROM, mail, message)
            except BaseException as e:
                Log.printlog('An error has occurred while sending a mail for warn about remote_platform', 6011)
    def notify_status_change(self, exp_id, job_name, prev_status, status, mail_to):
        message_text = self._generate_message_text(exp_id, job_name, prev_status, status)
        message = MIMEText(message_text)
        message['From'] = email.utils.formataddr(('Autosubmit', self.config.MAIL_FROM))
        message['Subject'] = f'[Autosubmit] The job {job_name} status has changed to {str(status)}'
        message['Date'] = email.utils.formatdate(localtime=True)
        for mail in mail_to:
            message['To'] = email.utils.formataddr((mail, mail))
            try:
                self._send_mail(self.config.MAIL_FROM, mail, message)
            except BaseException as e:
                Log.printlog('Trace:{0}\nAn error has occurred while sending a mail for the job {0}'.format(e,job_name), 6011)

    def _send_mail(self, mail_from, mail_to, message):
        server = smtplib.SMTP(self.config.SMTP_SERVER,timeout=60)
        server.sendmail(mail_from, mail_to, message.as_string())
        server.quit()

    @staticmethod
    def _generate_message_text(exp_id, job_name, prev_status, status):
        return f'Autosubmit notification\n' \
               f'-------------------------\n\n' \
                f'Experiment id:  {str(exp_id)} \n\n' \
                + f'Job name: {str(job_name)} \n\n' \
                f'The job status has changed from: {str(prev_status)} to {str(status)} \n\n\n\n\n' \
                f'INFO: This message was auto generated by Autosubmit, '\
                f'remember that you can disable these messages on Autosubmit config file. \n'

    @staticmethod
    def _generate_message_experiment_status(exp_id, platform=""):
        return f'Autosubmit notification: Remote Platform issues\n' \
               f'-------------------------\n\n' \
                f'Experiment id:{str(exp_id)} \n\n' \
                + f'Platform affected:{str(platform.name)} using as host:{str(platform.host)} \n\n' \
                f'[WARN] Autosubmit encountered an issue with an remote_platform.\n It will resume itself, whenever is possible\n If issue persist, you can change the host IP or put multiple hosts in the platform.yml' + '\n\n\n\n\n' \
                f'INFO: This message was auto generated by Autosubmit, '\
                f'remember that you can disable these messages on Autosubmit config file. \n'